from a2a.server.events import EventQueue


from a2a_acl.agent.acl_agent import ACLAgentExecutor
from a2a_acl.content_codecs.common import atom_codec_id

from a2a_acl.protocol_extension.acl_message import ACLMessage

from a2a_acl.interface import asi_parser
from a2a_acl.interface.interface import ACLAgentCard


my_card = asi_parser.read_file("failing.asi").add_codecs([atom_codec_id])


class PingableAgentExecutor(ACLAgentExecutor):

    def __init__(self, card: ACLAgentCard, url):
        super().__init__(agentcard=card, my_url=url)
        print(
            "Warning: launching a pure python A2A agent, unable to check .asi interface against python body."
        )

    async def execute_achieve(
        self,
        m: ACLMessage,
        output_event_queue: EventQueue,
    ):
        print("(Received a message: " + str(m) + ")")
        match m.content:
            case "ping":
                print("Received a ping from " + m.sender)
                raise Exception("bad")



import ast

import context
import asyncio

from a2a.server.events import EventQueue
from a2a.server.tasks import TaskUpdater
from a2a.types import TaskState, Part, TextPart
from a2a.utils import new_task, new_agent_text_message

from a2a_acl.agent.acl_agent import ACLAgentExecutor
from a2a_acl.content_codecs.common import natural_language_id, atom_codec_id

from a2a_acl.protocol_extension.acl_message import ACLIncomingMessage
from a2a_acl.a2a_utils.send_message import sync_reply

from a2a_acl.interface import asi_parser
from a2a_acl.protocol_extension.send_acl_message import spawn_send_acl_message
from sample_agents.llm_refactoring_advice.prompts import ask_llm_for_refactoring

my_card = asi_parser.read_file("advisor.asi").add_codecs([natural_language_id])


def decode_python(llm_answer: str):

    tag1 = "```python"
    tag2 = "```"
    if tag1 in llm_answer:
        (pre, tmp) = llm_answer.split(tag1)
        if tag2 in tmp:
            (code, post) = tmp.split(tag2)
            ast.parse(code)  # raises an exception on some syntax errors
            return (pre, code, post)
    raise ValueError(llm_answer)


class LLMAdvisor(ACLAgentExecutor):

    def __init__(self, url):
        super().__init__(agentcard=my_card, my_url=url)
        print(
            "Warning: launching a pure python A2A agent,"
            " unable to check .asi interface against python body."
        )

        self.spec = "A function that does nothing."

    async def execute_tell(
        self,
        m: ACLIncomingMessage,
        output_event_queue: EventQueue,
    ):
        self.spec = m.content
        await sync_reply(output_event_queue, "ack")
        print("Received a new specification to consider.")

    async def execute_on_upload(
        self,
        m: ACLIncomingMessage,
        output_event_queue: EventQueue,
    ):

        print("(creating a new task)")
        task = new_task(m.origin.message)
        await output_event_queue.enqueue_event(task)
        updater = TaskUpdater(output_event_queue, task.id, task.context_id)

        await asyncio.sleep(1)

        await updater.start_work()

        await asyncio.sleep(1)

        print("(consult llm)")
        res = ask_llm_for_refactoring(self.spec, m.content, context.mock_flag)
        try:
            (a, code, c) = decode_python(res)
            await updater.update_status(
                TaskState.working,
                new_agent_text_message(
                    "LLM first comment:" + a, context_id=m.origin.context_id
                ),
            )
            await updater.update_status(
                TaskState.working,
                new_agent_text_message(
                    "LLM second comment (possibly truncated):" + c,
                    context_id=m.origin.context_id,
                ),
            )

            await asyncio.sleep(1)

            await updater.add_artifact(
                [Part(root=TextPart(text=code))], name="Final response."
            )

            await asyncio.sleep(1)

            await updater.complete()

            return

        except SyntaxError:
            e = "Failed to generate syntactically correct code."
            print("ERROR: " + e)
            await self.report_error(updater, m.sender, e)

        except Exception as e:
            print("ERROR: " + str(type(e)))
            await self.report_error(updater, m.sender, "unknown reason")

    async def report_error(self, updater, to, m: str):
        # synchronous
        await updater.update_status(TaskState.failed)

        # asynchronous
        if to != "":
            spawn_send_acl_message(to, "failure", m, self.my_url, natural_language_id)



from dataclasses import dataclass

from a2a_acl.protocol_extension.acl_message import Illocution


@dataclass
class SkillRequest:
    """Description of skills used to filter skills."""

    declaration_kind: Illocution
    functor: str
    arity: int


@dataclass
class SkillDeclaration:
    """A skill as declared in a .asi interface file"""

    declaration_kind: Illocution
    functor: str
    arity: int
    doc: str

    def conforms_to(self, intf: SkillRequest) -> bool:
        return (
            self.declaration_kind == intf.declaration_kind
            and self.functor == intf.functor
            and self.arity == intf.arity
        )


@dataclass
class ACLAgentCard:
    """An agent card as declared in a .asi interface file."""

    name: str
    doc: str
    skills: list[SkillDeclaration]
    supported_codecs: list[str]

    def has_declared_skill(self, r: SkillRequest) -> bool:
        for skill in self.skills:
            if skill.conforms_to(r):
                return True
        return False

    def has_declared_skills(self, s: list[SkillRequest]) -> bool:
        # forall
        for skill in s:
            if not self.has_declared_skill(skill):
                return False
        return True

    def add_codecs(self, codecs):
        for c in codecs:
            self.supported_codecs.append(c)
        return self



import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../..")))

advisor_port = 9989

# Change this flag to use the LLM instead of a mock
mock_flag = True
MOCK_FILE = "mock.txt"
# MOCK_FILE = "mock_error.txt"

# Change this to change the LLM
llm_model = "mistral/mistral-small-latest"
# llm_model = "gpt-4o-mini"

max_token = 100



from a2a.server.events import EventQueue

import a2a_acl
from a2a_acl.agent.acl_agent import ACLAgentExecutor
from a2a_acl.content_codecs.common import atom_codec_id

from a2a_acl.protocol_extension.acl_message import ACLMessage
from a2a_acl.a2a_utils.send_message import sync_reply

from a2a_acl.interface import asi_parser
from a2a_acl.interface.interface import ACLAgentCard


my_card = asi_parser.read_file("pingable.asi").add_codecs([atom_codec_id])


class PingableAgentExecutor(ACLAgentExecutor):

    def __init__(self, card: ACLAgentCard, url):
        super().__init__(agentcard=card, my_url=url)
        print(
            "Warning: launching a pure python A2A agent, unable to check .asi interface against python body."
        )

    async def execute_achieve(
        self,
        m: ACLMessage,
        output_event_queue: EventQueue,
    ):
        print("(Received a message: " + str(m) + ")")
        match m.content:
            case "ping":
                print("Received a ping from " + m.sender)
                await sync_reply(output_event_queue, "Ack: Achieve received")
                await output_event_queue.close()
                self.send_message(
                    m.sender,
                    "tell",
                    "pong",
                    a2a_acl.content_codecs.common.atom_codec_id,
                )
                return



import os
from mistralai import Mistral

llm_api_key = os.environ["MISTRAL_API_KEY"]
llm_model = "mistral-small-latest"

llm_client = Mistral(api_key=llm_api_key)


def log(m):
    print("[LOG] " + m)


def ask_llm(system_prompt, user_prompt):

    messages = [system_prompt, user_prompt]

    log("Full prompt: " + str(messages))

    chat_response = llm_client.chat.complete(
        model=llm_model,
        messages=messages,
        max_tokens=5,
    )

    response = chat_response.choices[0].message.content

    log(
        "I had an interaction with mistral with the prompt above.\n "
        + "Mistral gave me the following answer: "
        + response
    )
    return response



from dataclasses import dataclass
from enum import StrEnum

from a2a.server.agent_execution import RequestContext


class Illocution(StrEnum):
    """Illocutions that are used in messages.
    * TELL, ACHIEVE, ASK are basic illocutions inspired by KQML.
    * PROPOSE is suited for arbitration or consensus.
    * UPLOAD is suited for data transfer.
    * FAILURE is to be used to report a failure in operation.
    """

    TELL = "tell"
    ACHIEVE = "achieve"
    ASK = "ask"
    PROPOSE = "propose"
    UPLOAD = "upload"
    FAILURE = "failure"


@dataclass
class ACLMessage:
    illocution: Illocution
    content: str
    sender: str
    codec: str


@dataclass
class ACLIncomingMessage(ACLMessage):
    task_id: str
    origin: RequestContext



import context

import unittest
import uuid
from typing import Any

from a2a.server.agent_execution import RequestContext
from a2a.types import Message, Role, Part, TextPart, MessageSendParams

from a2a_acl.protocol_extension import MIME_TAG, message_codec
from a2a_acl.protocol_extension.message_codec import check_illocution_regexpr


URI = "https://gitlab.eclipse.org/eclipse-research-labs/mosaico-project/a2a-acl/"


class TestACLtoA2A(unittest.TestCase):
    def test_illocution_in_message(self):
        md: dict[str, Any] = {
            "illocution": "tell",
            "codec": "nl",
            "mediaType": MIME_TAG,
        }
        m = Message(
            role=Role.agent,
            message_id=uuid.uuid4().hex,
            parts=[Part(root=TextPart(text="I am ready."))],
            extensions=[URI],
            metadata={URI: md},
        )
        p = MessageSendParams(message=m)
        c = RequestContext(request=p)
        self.assertEqual(message_codec.acl_of_a2a(c).illocution, "tell")

    def test_illocution_regexpr(self):
        self.assertTrue(check_illocution_regexpr("tell"))
        self.assertTrue(check_illocution_regexpr("not-understood"))
        self.assertFalse(check_illocution_regexpr("not ready"))
        self.assertFalse(check_illocution_regexpr("not_ready"))
        self.assertFalse(check_illocution_regexpr("-not-ready"))
        self.assertFalse(check_illocution_regexpr("Ready"))
        self.assertFalse(check_illocution_regexpr(" ready"))
        self.assertFalse(check_illocution_regexpr(""))
        self.assertFalse(check_illocution_regexpr("ready1"))


if __name__ == "__main__":
    unittest.main()



import time

import a2a

import context

from a2a.types import (
    AgentCard,
)

from a2a_acl.a2a_utils.send_message import (
    extract_text_from_message,
    extract_text_from_task_artifact,
)
from a2a_acl.protocol_extension.send_acl_message import (
    spawn_send_acl_message,
)
from a2a_acl.a2a_utils.card_holder import download_card


from a2a_acl.utils.url import build_url

host = "127.0.0.1"


def handle_status_update(e, t):
    match type(t):
        case a2a.types.TaskStatusUpdateEvent:
            print("(status update: " + str(t.status.state) + ")")
        case _:
            print("(status update ignored)")


async def main() -> None:
    target_agent_url = build_url(host, context.port_validator)

    # Fetch Public Agent Card and Initialize Client
    target_agent_card: AgentCard | None = None

    try:
        target_agent_card = await download_card(target_agent_url)

    except Exception:
        print("Client failed to fetch the public agent card. Cannot continue.")
        exit(-1)

    spawn_send_acl_message(
        target_agent_card,
        "tell",
        "A function to reverse the order of the elements of a list.",
        "",
        "nl",
        message_callback=(
            lambda m: print("Message event: " + extract_text_from_message(m))
        ),
        artifact_update_callback=(
            lambda t: print(
                "Task artifact event: " + extract_text_from_task_artifact(t)
            )
        ),
        status_update_callback=handle_status_update,
    )
    print("[Tell-Message sent to validator agent.]")

    await asyncio.sleep(0.5)

    spawn_send_acl_message(
        target_agent_card,
        "propose",
        "The function should return an integer.",
        "",
        "nl",
        message_callback=(lambda m: print("Message: " + str(m))),
        artifact_update_callback=(
            lambda t: print("Task artifact: " + extract_text_from_task_artifact(t))
        ),
        status_update_callback=handle_status_update,
    )
    print("[Propose-Message sent to validator agent.]")


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())
    input("Press Enter to exit.\n")



from a2a_acl.content_codecs.python_codec import PythonCodec

"""Atom codec is for literals reduced to a word, such as 'ping', or 'build'. 
Two agents using this codec to dialogue can be implemented in different languages.
 In the Python implementation of BDI agents, we use the 'Python Codec' to implement the Atom codec. """

codec_object = PythonCodec()



import os
import sys

from a2a_acl.utils.url import build_url

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../..")))

the_host = "127.0.0.1"

generator_port = 9990
manager_port = 9981
validator_port = 9999
reference_agent_port = 9970

spec = "A function to compare two words."




from a2a.server.events import EventQueue

from a2a_acl.agent.acl_agent import ACLAgentExecutor
from a2a_acl.content_codecs.common import natural_language_id

from a2a_acl.protocol_extension.acl_message import ACLMessage, Illocution
from a2a_acl.a2a_utils.send_message import sync_reply

from a2a_acl.interface import asi_parser
from a2a_acl.interface.interface import ACLAgentCard


my_card = asi_parser.read_file("human_validator.asi").add_codecs([natural_language_id])


def ask_validation(ref: str, t: str):
    print("Reference: " + ref)
    print("Proposal: " + t)
    print("Do you validate this proposal ?")
    res = input("[Y/N/Q]\n")
    match res:
        case "Y" | "y":
            print("Validated. (" + res + ")")
            return "valid"
        case "Q" | "q":
            print("Quit. (" + res + ")")
            return "quit"
        case _:
            print("Not Validated. (" + res + ")")
            return "invalid"


class HumanValidatorAgentExecutor(ACLAgentExecutor):
    """An agent that receives validation requests and
    which delegates the decision to a human user."""

    def __init__(self, card: ACLAgentCard, url):
        super().__init__(agentcard=card, my_url=url)
        print(
            "Warning: launching a pure python A2A agent, unable to check .asi interface against python body."
        )
        self.add_message_handler(Illocution.PROPOSE, self.execute_propose)
        self.reference = ""

    async def execute_achieve(
        self,
        m: ACLMessage,
        output_event_queue: EventQueue,
    ):
        pass

    async def execute_tell(
        self,
        m: ACLMessage,
        output_event_queue: EventQueue,
    ):
        self.reference = m.content
        await sync_reply(output_event_queue, "ack")
        print("Received an update of the reference.")

    async def execute_propose(
        self,
        m: ACLMessage,
        output_event_queue: EventQueue,
    ):
        res = ask_validation(self.reference, m.content)
        await sync_reply(output_event_queue, res)
        return



import os

from a2a.types import (
    AgentCard,
)
from a2a.utils import new_agent_text_message

from a2a.server.events import EventQueue

from a2a_acl.agent.acl_agent import ACLAgentExecutor
from a2a_acl.agent.server_utils import run_server
from a2a_acl.protocol_extension.acl_message import ACLMessage, Illocution
import a2a_acl.protocol_extension.send_acl_message as send_acl_message
from a2a_acl.a2a_utils.card_holder import download_card
from a2a_acl.content_codecs.common import atom_codec_id
from a2a_acl.interface.interface import (
    SkillDeclaration,
    ACLAgentCard,
)
from a2a_acl.utils.url import build_url

my_host = "127.0.0.1"
my_port = 9998
my_url = build_url(my_host, my_port)
other_agent_url = build_url(my_host, 9999)

my_skill = SkillDeclaration(
    Illocution.TELL, "pong", 0, "Receive a pong answer to a ping request."
)

my_card = ACLAgentCard("Client Agent", "A client agent", [my_skill], [atom_codec_id])


class ClientAgentExecutor(ACLAgentExecutor):
    def __init__(self, url):
        super().__init__(my_card, url)
        self.pong_received = 0

    async def execute_tell(
        self,
        m: ACLMessage,
        output_event_queue: EventQueue,
    ) -> None:
        if m.content == "pong":
            print("pong received")
            self.pong_received += 1
        if self.pong_received == 2:
            print("TEST OK")
        elif self.pong_received > 2:
            print("TEST KO)")
        await output_event_queue.enqueue_event(new_agent_text_message("Tell received"))


def on_error():
    print("KO : failure while talking to the agent.")
    os._exit(-1)


async def main() -> None:

    # 1) start an a2a server
    run_server(ClientAgentExecutor(my_url), my_host, my_port)

    # 2) query the other a2a agent

    # Fetch Public Agent Card and Initialize Client
    target_agent_card: AgentCard | None = None
    try:
        target_agent_card = await download_card(other_agent_url)

    except Exception:
        print("Client failed to fetch the public agent card. Cannot continue.")
        exit(-1)

    print(
        "Going to send 2 achieve/ping messages, expect to receive two tell/pong messages."
    )
    # First message
    send_acl_message.spawn_send_acl_message(
        target_agent_card,
        "achieve",
        "ping",
        my_url,
        atom_codec_id,
        error_callback=on_error,
    )

    await asyncio.sleep(1)

    # Second message
    send_acl_message.spawn_send_acl_message(
        target_agent_card,
        "achieve",
        "ping",
        my_url,
        atom_codec_id,
        error_callback=on_error,
    )


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())

    input("Press Enter to exit\n")



def neutralize_str(s: str) -> str:
    """Add quotes to a string."""
    return '"' + s + '"'


def clear(s: str):
    """Remove quotes from a string."""
    if s.startswith('"') and s.endswith('"'):
        return s.removeprefix('"').removesuffix('"')
    elif s.startswith("'") and s.endswith("'"):
        return s.removeprefix("'").removesuffix("'")
    else:
        return s



from a2a.types import AgentCard

from a2a_acl.protocol_extension.acl_message import Illocution
from a2a_acl.interface.skill_conversion import bdi_skill_of_a2a_skill
from a2a_acl.translator_nl_to_acl.mistral_config import ask_llm


class LLMError(Exception):
    pass


system_prompt_message = {
    "role": "system",
    "content": "Your task is to translate a human request into an ACL message",
}


def build_user_prompt_message(agent_type: str, catalog: str, user_request: str):
    return {
        "role": "user",
        "content": (
            "Your task is to translate a human request into an ACL message for the following agent: "
            + agent_type
            + " The possible achievements are described in the following list. [BEGIN LIST OF ACHIEVEMENTS] "
            + catalog
            + "[END OF LIST OF ACHIEVEMENTS]"
            + " Here is the sentence to translate. [BEGIN] "
            + user_request
            + "[END] "
            + "Respond with only one achievement. Do not include the illocution in your answer. "
            + "For example you can answer 'do_dig' if that achievement has arity 0, or 'do_wait(300)' if arity is 1."
        ),
    }


def handled_illocutions(card: AgentCard) -> set:
    res = set()
    for s in card.skills:
        res.add(bdi_skill_of_a2a_skill(s).declaration_kind)
    return res


def doc_illoc(i: Illocution) -> str:
    prefix = str(i) + " ("
    suffix = ")"
    match i:
        case Illocution.TELL:
            return prefix + "to give an information" + suffix
        case Illocution.ACHIEVE:
            return prefix + "to request an action" + suffix
        case Illocution.ASK:
            return prefix + "to ask for some information" + suffix


def select_illocution(card: AgentCard, request: str) -> Illocution:
    s = handled_illocutions(card)
    l = [e for e in s]
    n = len(l)
    assert n > 0
    if n == 1:
        return l[0]
    else:
        system_prompt_message = {
            "role": "system",
            "content": "Your task is to extract an ACL illocution from a human request.",
        }
        user_prompt = {
            "role": "user",
            "content": (
                "Your task is to extract an ACL illocution from a human request. "
                + " The possible ACL illocutions are: "
                + str([doc_illoc(i) for i in l])
                + " Here is the human request. [BEGIN] "
                + request
                + "[END] "
                + "Respond with only one illocution. "
            ),
        }
        r = ask_llm(system_prompt=system_prompt_message, user_prompt=user_prompt)
        return r


def clean_skill(s: AgentSkill) -> str:
    return (
        "Goal: " + s.id + ", " + s.description + ", Examples: " + str(s.examples) + "."
    )


def clean(skills) -> str:
    accu = "["
    for s in skills:
        accu += clean_skill(s) + " "
    accu += "]"
    return accu


def translate(card: AgentCard, request: str) -> str:
    try:
        illoc = select_illocution(card=card, request=request)
    except Exception:
        raise LLMError()
    try:
        user_prompt = build_user_prompt_message(
            str(card.description), clean(card.skills), request
        )
        res = ask_llm(system_prompt=system_prompt_message, user_prompt=user_prompt)
    except Exception:
        raise LLMError()
    return (illoc, res)



import asyncio
import uuid
from threading import Thread
from typing import Callable, Any

import a2a
import httpx
from a2a.client import ClientFactory, ClientConfig
from a2a.server.events import EventQueue
from a2a.types import (
    Message,
    Role,
    Part,
    TextPart,
    AgentCard,
    Task,
    PushNotificationConfig,
    TaskStatusUpdateEvent,
)

from a2a.utils import new_agent_text_message

from a2a_acl.a2a_utils.card_holder import download_card


def build_message(content: str) -> Message:
    return Message(
        role=Role.agent,
        message_id=uuid.uuid4().hex,
        parts=[Part(root=TextPart(text=content))],
    )


class SendFailureException(Exception):
    pass


def extract_text_from_message(m: Message) -> str:
    return m.parts[0].root.text


def extract_text_from_task_artifact(t: Task) -> str:
    res = str(t.status.state)
    if t.artifacts:
        a = t.artifacts[-1]
        for p in a.parts:
            if p.root.kind == "text":
                res = p.root.text
    return res


def default_message_handler(m: Message) -> None:
    print("(Message event received: " + extract_text_from_message(m) + ")")


def default_artifact_update_handler(t: Task) -> None:
    print("(Artifact event received: " + extract_text_from_task_artifact(t) + ")")


def default_status_update_handler(
    t: Task, e: TaskStatusUpdateEvent, message_id: str
) -> None:
    print(
        "(Status update event received for task "
        + t.id
        + " initiated by message "
        + message_id
        + " : "
        + str(e.status.state)
        + ")"
    )


def default_error_callback():
    print(
        "Error while talking with agent (you can specify a callback to handle this error)."
    )


# derived from https://github.com/a2aproject/a2a-samples/blob/main/samples/python/agents/number_guessing_game/utils/protocol_wrappers.py
async def send_message(
    destination: AgentCard | str,
    msg: Message,
    sender: str,
    message_callback: Callable[[Message], None] = default_message_handler,
    artifact_update_callback: Callable[[Task], None] = default_artifact_update_handler,
    status_update_callback: Callable[
        [Any, a2a.types.TaskStatusUpdateEvent, str], None
    ] = default_status_update_handler,
):
    if isinstance(destination, str):
        try:
            destination = await download_card(destination)
        except Exception:
            print("Error: Cannot get agent card. Send failed.")
            raise SendFailureException

    else:
        assert isinstance(destination, AgentCard)

    _client_factory = ClientFactory(
        ClientConfig(
            push_notification_configs=[PushNotificationConfig(url=sender)],
            httpx_client=httpx.AsyncClient(timeout=httpx.Timeout(timeout=30)),
            streaming=destination.capabilities.streaming,
        )
    )
    client = _client_factory.create(destination)

    try:
        # Warning : concurrent async generators work badly in python
        # -> two concurrent message sending should be in two separate threads
        # (see spawn_send_message below)
        async for event in client.send_message(msg):
            if isinstance(event, Message):
                message_callback(event)

            else:
                assert isinstance(event, tuple)
                (t, opt) = event
                assert isinstance(t, Task)
                match type(opt):
                    case a2a.types.TaskStatusUpdateEvent:
                        status_update_callback(t, opt, msg.message_id)
                    case a2a.types.TaskArtifactUpdateEvent:
                        artifact_update_callback(t)
                    case None:
                        artifact_update_callback(t)
        print("End of sending")

    except a2a.client.errors.A2AClientTimeoutError:
        print(
            "Warning: no synchronous reply before timeout. Some information might be lost."
        )
    except a2a.client.errors.A2AClientHTTPError as e:
        print("Send failed (" + e.message + ")")
        raise SendFailureException

    except Exception as e:
        print("Send failed. (" + str(type(e)) + ", " + str(e) + ")")
        raise SendFailureException


# hard reference to task that must not be garbage collected.
launched_tasks = []
# Note that if we never delete tasks from this list, we create a memory leak. FIXME


class SendMessageThread(Thread):

    def __init__(
        self,
        dest: AgentCard | str,
        msg: Message,
        sender: str,
        error_callback,
        message_callback=default_message_handler,
        artifact_update_callback=default_artifact_update_handler,
        status_update_callback=default_status_update_handler,
    ):
        super().__init__()
        self.dest = dest
        self.msg = msg
        self.reply_to = sender
        self.message_callback = message_callback
        self.artifact_update_callback = artifact_update_callback
        self.status_update_callback = status_update_callback
        self.error_callback = error_callback

    def run(self):
        try:
            loop = asyncio.new_event_loop()
            t = loop.create_task(
                send_message(
                    self.dest,
                    self.msg,
                    self.reply_to,
                    message_callback=self.message_callback,
                    artifact_update_callback=self.artifact_update_callback,
                    status_update_callback=self.status_update_callback,
                )
            )
            launched_tasks.append(t)
            # TRY-ME : t.add_done_callback(launched_tasks.remove)
            loop.run_until_complete(t)
            print("(end of send thread)")

        except SendFailureException:
            print("Send failed.")
            self.error_callback()


def spawn_send_message(
    dest: AgentCard | str,
    content: str,
    sender: str,
    message_callback=default_message_handler,
    task_callback=default_artifact_update_handler,
    status_update_callback=default_status_update_handler,
    error_callback=default_error_callback,
) -> None:
    msg = build_message(content)
    t = SendMessageThread(
        dest,
        msg,
        sender,
        error_callback,
        message_callback,
        task_callback,
        status_update_callback,
    )
    t.start()


async def sync_reply(output_event_queue: EventQueue, r: str):
    # Warning : according to A2A spec v1.0rc1, you can reply with only 1 Message then close, or a stream of Task.
    await output_event_queue.enqueue_event(new_agent_text_message(r))



import ast

from a2a_acl.content_codecs.common import BadFormat, Codec, python_codec_id


class PythonCodec(Codec):

    def __init__(self):
        super().__init__(python_codec_id)

    def extract_entry_point(self, s: str):
        (a, b) = self.decode(s)
        return a

    def decode(self, s: str):
        try:
            if "(" in s:
                (a, b) = s.split("(", 1)
                c = b.removesuffix(")")
                d = ast.literal_eval(c)
                return (a, d)
            else:
                return (s, None)
        except Exception:
            print("Failed to decode " + s)
            raise BadFormat


codec_object = PythonCodec()



import threading

import uvicorn
from a2a.server.agent_execution import AgentExecutor
from a2a.server.apps import A2AStarletteApplication
from a2a.server.request_handlers import DefaultRequestHandler
from a2a.server.tasks import InMemoryTaskStore
from a2a.types import AgentCard

from a2a_acl.agent.acl_agent import ACLAgentExecutor
from a2a_acl.interface.card_conversion import a2a_card_from_bdi_card
from a2a_acl.utils.url import build_url


def run_a2a_server(
    card: AgentCard, agent_executor: AgentExecutor, host: str, port: int
):

    request_handler = DefaultRequestHandler(
        agent_executor=agent_executor,
        task_store=InMemoryTaskStore(),
    )

    server = A2AStarletteApplication(
        agent_card=card,
        http_handler=request_handler,
    )

    def start():
        uvicorn.run(server.build(), host=host, port=port)

    threading.Thread(target=start).start()
    print("-running a2a-server for " + card.name + " agent-")


def run_server(agent_executor: ACLAgentExecutor, host: str, port: int):
    agent_card = a2a_card_from_bdi_card(agent_executor.card, build_url(host, port))
    run_a2a_server(agent_card, agent_executor, host, port)



from a2a.server.events import EventQueue

from a2a_acl.agent.acl_agent import ACLAgentExecutor
from a2a_acl.content_codecs.common import natural_language_id

from a2a_acl.protocol_extension.acl_message import ACLMessage, Illocution
from a2a_acl.a2a_utils.send_message import sync_reply

from a2a_acl.interface import asi_parser
from a2a_acl.interface.interface import ACLAgentCard

import prompts
from sample_agents.llm_validator_mistral_api.prompts import ask_llm_for_correctness

my_card = asi_parser.read_file("validator.asi").add_codecs([natural_language_id])


class ValidatorAgentExecutor(ACLAgentExecutor):

    def __init__(self, card: ACLAgentCard, url):
        super().__init__(agentcard=card, my_url=url)
        print(
            "Warning: launching a pure python A2A agent, unable to check .asi interface against python body."
        )
        self.add_message_handler(Illocution.PROPOSE, self.execute_propose)
        self.spec = "A function to compare reverse the order of the elements in a list."

    async def execute_achieve(
        self,
        m: ACLMessage,
        output_event_queue: EventQueue,
    ):
        pass

    async def execute_tell(
        self,
        m: ACLMessage,
        output_event_queue: EventQueue,
    ):
        self.spec = m.content
        await sync_reply(output_event_queue, "ack")
        print("Received an update of the specification.")

    async def execute_ask(
        self,
        m: ACLMessage,
        output_event_queue: EventQueue,
    ):
        pass

    async def execute_propose(
        self,
        m: ACLMessage,
        output_event_queue: EventQueue,
    ):

        res = ask_llm_for_correctness(
            self.spec,
            m.content,
        )
        await sync_reply(output_event_queue, "valid" if res else "invalid")
        return



import context
from a2a.types import (
    AgentCard,
)
from a2a.utils import new_agent_text_message

from a2a.server.events import EventQueue
from a2a_acl.agent.acl_agent import ACLAgentExecutor
from a2a_acl.agent.server_utils import run_server
from a2a_acl.content_codecs.common import atom_codec_id
from a2a_acl.protocol_extension.acl_message import ACLMessage, Illocution
from a2a_acl.protocol_extension.send_acl_message import spawn_send_acl_message
from a2a_acl.a2a_utils.card_holder import download_card
from a2a_acl.interface.interface import SkillDeclaration, ACLAgentCard
from a2a_acl.utils.url import build_url


my_skill = SkillDeclaration(
    Illocution.TELL, "pong", 0, "Receive a pong answer to a ping request."
)

my_card = ACLAgentCard("Client Agent", "A client agent", [my_skill], [atom_codec_id])

my_host = "127.0.0.1"

my_url = build_url(my_host, context.pinger_port)

other_agent_url = build_url(my_host, context.pingable_port)


class ClientAgentExecutor(ACLAgentExecutor):
    def __init__(self, my_url):
        super().__init__(my_card, my_url)
        self.pong_received = 0

    async def execute_tell(
        self,
        m: ACLMessage,
        output_event_queue: EventQueue,
    ) -> None:
        print("ACL message received.")
        if m.content == "pong":
            print("tell/pong received")
            self.pong_received += 1
        else:
            print("Other message received")
        if self.pong_received == 1:
            print("TEST OK")
        elif self.pong_received > 1:
            print("TEST KO)")
        await output_event_queue.enqueue_event(new_agent_text_message("Tell received"))


async def main() -> None:

    # 1) start an a2a server
    run_server(ClientAgentExecutor(my_url), my_host, context.pinger_port)

    # 2) query the other a2a agent

    # Fetch Public Agent Card
    target_agent_card: AgentCard | None = None
    try:
        target_agent_card = await download_card(other_agent_url)
        print("Card downloaded.")

    except Exception:
        print("Client failed to fetch the public agent card. Cannot continue.")
        exit(-1)

    # Send message (achieve)
    spawn_send_acl_message(target_agent_card, "achieve", "ping", my_url, atom_codec_id)
    print("Message sent.")


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())

    input("Press Enter to exit.")



import os
import litellm

llm_model = "mistral/mistral-small-latest"

os.environ["LANGFUSE_OTEL_HOST"] = "https://cloud.langfuse.com"
litellm.callbacks = ["langfuse_otel"]


def log(m):
    print("[LOG] " + m)


def filter(t: str):
    t = t.lower()
    if t.startswith("true"):
        return True
    elif t.startswith("false"):
        return False
    else:
        return None


def ask_llm_for_correctness(spec: str, req: str) -> bool:
    try:
        chat_response = litellm.completion(
            model=llm_model,
            messages=[
                {
                    "role": "system",
                    "content": "Given a specification of a system, and a requirement,"
                    " evaluate if that requirements is correct with respect to the given specification."
                    + " Just answer True or False if you are sure. Just answer Unclear if you miss some information to decide. Don't explain.",
                },
                {
                    "role": "user",
                    "content": "Specification: "
                    + spec
                    + "(end of the specification) Requirement: "
                    + req
                    + "(end of requirement).",
                },
            ],
            max_tokens=5,
            temperature=1.2,
        )
        res = chat_response.choices[0].message.content
        log(
            "I had an interaction with "
            + llm_model
            + " to evaluate the correctness of a requirement."
        )
        log("I gave the following spec: " + spec)
        log("I also gave the following requirement: " + req)
        log("The LLM gave me the following answer: " + res)
        b = filter(res)
        return b
    except Exception as e:
        print(str(e))
        raise Exception("LLM failure: " + str(type(e))) from e



import context

from a2a.types import (
    AgentCard,
)

from a2a_acl.content_codecs.common import atom_codec_id
from a2a_acl.protocol_extension.send_acl_message import spawn_send_acl_message
from a2a_acl.a2a_utils.card_holder import download_card


from a2a_acl.utils.url import build_url

host = "127.0.0.1"


async def main() -> None:
    target_agent_url = build_url(host, context.port_pinger)

    # Fetch Public Agent Card and Initialize Client
    target_agent_card: AgentCard | None = None

    try:
        target_agent_card = await download_card(target_agent_url)

    except Exception:
        print("Client failed to fetch the public agent card. Cannot continue.")
        exit(-1)

    spawn_send_acl_message(target_agent_card, "achieve", "do_ping", "", atom_codec_id)
    print("Message sent to pinger agent.")


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())
    input("Press Enter to exit.\n")



from a2a.types import AgentSkill

from a2a_acl.protocol_extension.acl_message import Illocution
from a2a_acl.interface.interface import SkillDeclaration


def generate_param_example(i: int) -> str:
    if i is None:
        return ""
    else:
        return "(" + ("0," * i).removesuffix(",") + ")"


def pretty_print_example(s: SkillDeclaration) -> str:
    """Example of output: '(achieve,share_secret(0))'"""
    a = s.arity
    ex_param = "" if a == 0 else generate_param_example(a)
    ex = s.functor + ex_param
    return "(" + str(s.declaration_kind) + "," + ex + ")"


def parse_example(s: str) -> tuple[str, str]:
    """
    Example of input: '(achieve,share_secret(0))'
    Input encoded by pretty_print_example above.
    """
    s2 = s.removeprefix("(").removesuffix(")")
    (a, b) = s2.split(",")
    f = b.split("(")[0]
    return a, f


def encode_arity(a: int) -> str:
    if a is None:
        return ""
    else:
        return "(needs " + str(a) + " parameter" + ("s" if a > 1 else "") + ")"


def decode_arity(s: str) -> int:
    """
    Example of input: 'send a secret (needs 1 parameter)'
    Inputs are encoded by encode_arity above.
    Return 1 in this example.
    """
    l = s.split(" ")
    s1 = l[-2]
    return int(s1)


def a2a_skill_of_bdi_skill(s: SkillDeclaration) -> AgentSkill:
    return AgentSkill(
        id=s.functor,
        name=s.functor + " (from ACL agent)",
        description=s.doc + " " + encode_arity(s.arity),
        tags=[s.functor],
        examples=[pretty_print_example(s)],
    )


def bdi_skill_of_a2a_skill(s: AgentSkill) -> SkillDeclaration:
    """
    Exemple of skill for a ACL agent :
    AgentSkill(
        description='send a secret (needs 1 parameter)',
        examples=['(achieve,share_secret(0))'],
        id='share_secret',
        input_modes=None,
        name='share_secret (from ACL agent)',
        output_modes=None,
        security=None,
        tags=['share_secret']
        )
    """
    (a, f) = parse_example(s.examples[0])
    return SkillDeclaration(
        declaration_kind=Illocution(a),
        functor=f,
        arity=decode_arity(s.description),
        doc=s.description,
    )



from a2a.server.agent_execution import AgentExecutor, RequestContext
from a2a.server.events import EventQueue
from a2a.types import AgentCard

import a2a_acl.content_codecs.common as common
from a2a_acl.protocol_extension.message_codec import acl_of_a2a
from a2a_acl.protocol_extension.acl_message import Illocution, ACLIncomingMessage
from a2a_acl.protocol_extension.send_acl_message import (
    spawn_send_acl_message,
    default_message_handler,
    default_artifact_update_handler,
)
from a2a_acl.content_codecs.common import Codec
from a2a_acl.interface.interface import ACLAgentCard

import a2a_acl.content_codecs.python_codec as python_codec
import a2a_acl.content_codecs.atom_codec as atom_codec


class ACLAgentExecutor(AgentExecutor):

    def __init__(self, agentcard: ACLAgentCard, my_url: str):
        self.public_entry_points = [
            (s.declaration_kind, s.functor) for s in agentcard.skills
        ]
        self.codecs = agentcard.supported_codecs
        self.print_entry_points()
        self.print_codecs()
        self.card = agentcard
        self.my_url = my_url
        self.codec_objects = {
            common.python_codec_id: python_codec.codec_object,
            common.atom_codec_id: atom_codec.codec_object,
        }
        self.other_message_handlers = {}

    def print_entry_points(self):
        print("Public entry points:")
        for entry_point in self.public_entry_points:
            print(" " + str(entry_point))

    def print_codecs(self):
        print("Public codecs:")
        for codec in self.codecs:
            print(" " + str(codec))

    def get_codec_object(self, codec_name: str) -> Codec:
        if codec_name in self.codec_objects:
            return self.codec_objects[codec_name]
        else:
            print("Codec " + codec_name + " not supported.")
            raise NotImplementedError

    def firewall_accept_entry_point(self, m: ACLIncomingMessage) -> bool:
        ep = (
            "*"
            if m.codec == common.natural_language_id
            else self.get_codec_object(m.codec).extract_entry_point(m.content)
        )
        if (m.illocution, ep) not in self.public_entry_points:
            print("Received a message on an entry point which is not public: " + ep)
            return False
        else:
            return True

    def firewall_accept_upload(self):
        return (Illocution.UPLOAD, "*") in self.public_entry_points

    def firewall_accept_failure(self):
        return (Illocution.FAILURE, "*") in self.public_entry_points

    def codec_accept(self, m: ACLIncomingMessage) -> bool:
        if m.codec in self.codecs:
            return True
        else:
            print("Codec not supported : " + m.codec)
            return False

    async def execute(self, context: RequestContext, event_queue: EventQueue) -> None:
        m: ACLIncomingMessage = acl_of_a2a(context)

        match m.illocution:

            case Illocution.TELL:
                if self.codec_accept(m) and self.firewall_accept_entry_point(m):
                    await self.execute_tell(m, event_queue)
                else:
                    print("Message not handled.")

            case Illocution.ACHIEVE:
                if self.codec_accept(m) and self.firewall_accept_entry_point(m):
                    await self.execute_achieve(m, event_queue)
                else:
                    print("Message not handled.")

            case Illocution.ASK:
                if self.codec_accept(m) and self.firewall_accept_entry_point(m):
                    await self.execute_ask(m, event_queue)
                else:
                    print("Message not handled.")

            case Illocution.UPLOAD:
                """Upload messages contain data that is not parsed by the firewall
                (no codec, no entry-point)."""
                if self.firewall_accept_upload():
                    await self.execute_on_upload(m, event_queue)
                else:
                    print("Message not handled.")

            case Illocution.FAILURE:
                """Failure messages are not parsed by the firewall
                (no codec, no entry-point)."""
                if self.firewall_accept_failure():
                    await self.execute_on_failure(m, event_queue)
                else:
                    print("Message not handled.")

            case _:
                if (
                    self.codec_accept(m)
                    and self.firewall_accept_entry_point(m)
                    and m.illocution in self.other_message_handlers
                ):
                    f = self.other_message_handlers[m.illocution]
                    await f(m, event_queue)
                else:
                    print("Unknown illocution: " + m.illocution)
                    print("Message not handled.")

    # Override-me
    async def execute_tell(
        self, m: ACLIncomingMessage, event_queue: EventQueue
    ) -> None:
        pass

    # Override-me
    async def execute_achieve(
        self, m: ACLIncomingMessage, event_queue: EventQueue
    ) -> None:
        pass

    # Override-me
    async def execute_ask(self, m: ACLIncomingMessage, event_queue: EventQueue) -> None:
        pass

    # Override-me
    async def execute_on_upload(
        self, m: ACLIncomingMessage, event_queue: EventQueue
    ) -> None:
        pass

    # Override-me
    async def execute_on_failure(
        self, m: ACLIncomingMessage, event_queue: EventQueue
    ) -> None:
        pass

    def add_message_handler(self, illoc, f):
        self.other_message_handlers[illoc] = f

    def check_public_entry_point(self, m: ACLIncomingMessage, codec: Codec):
        entry_point = codec.extract_entry_point(m.content)
        return entry_point in self.public_entry_points

    async def cancel(self, context: RequestContext, event_queue: EventQueue) -> None:
        raise Exception("Cancel not supported.")

    def send_message(
        self,
        dest: AgentCard | str,
        illocution: str,
        content: str,
        codec: str,
        message_processor=default_message_handler,
        task_processor=default_artifact_update_handler,
    ):
        spawn_send_acl_message(
            dest,
            illocution,
            content,
            self.my_url,
            codec,
            message_processor,
            task_processor,
        )



import os
import context
from context import llm_model
import litellm

os.environ["LANGFUSE_OTEL_HOST"] = "https://cloud.langfuse.com"
litellm.callbacks = ["langfuse_otel"]


def log(m):
    print("[LOG] " + m)


def read_mock():
    return open(context.MOCK_FILE, "r").read()


def ask_llm_for_refactoring(spec: str, code: str, mock: bool = False) -> str:
    try:
        chat_response = litellm.completion(
            model=llm_model,
            messages=[
                {
                    "role": "system",
                    "content": "Given a specification of a system, and a source code,"
                    " propose an improvement of the code that does not change its behavior. First give the proposed code, then explain shortly.",
                },
                {
                    "role": "user",
                    "content": "Specification: "
                    + spec
                    + "(end of the specification) Source code: "
                    + code
                    + "(end of code).",
                },
            ],
            max_tokens=context.max_token,
            mock_response=(read_mock() if mock else None),
        )
        res = chat_response.choices[0].message.content
        log("I had an interaction with " + llm_model + " to suggest improvements.")
        log("I gave the following spec: " + spec)
        log("I also gave the following code: " + code)
        log("The LLM gave me the following answer: " + res)
        return res
    except Exception as e:
        print(str(e))
        raise Exception("LLM failure: " + str(type(e))) from e



import context
from a2a_acl.agent.server_utils import run_server

from a2a_acl.utils.url import build_url

from reference_agent import ReferenceAgentExecutor, my_card

my_url = build_url(context.the_host, context.reference_agent_port)


async def main() -> None:
    run_server(
        ReferenceAgentExecutor(my_card, my_url),
        context.the_host,
        context.reference_agent_port,
    )


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())



import asyncio

from a2a.server.events import EventQueue
from a2a.server.tasks import TaskUpdater
from a2a.types import TaskState, Part, TextPart
from a2a.utils import new_task, new_agent_text_message


from a2a_acl.agent.acl_agent import ACLAgentExecutor
from a2a_acl.content_codecs.common import natural_language_id

from a2a_acl.protocol_extension.acl_message import (
    ACLMessage,
    Illocution,
    ACLIncomingMessage,
)
from a2a_acl.a2a_utils.send_message import sync_reply

from a2a_acl.interface import asi_parser


from sample_agents.llm_req_validator_2.prompts import ask_llm_for_correctness

my_card = asi_parser.read_file("validator.asi").add_codecs([natural_language_id])


class LLMValidator(ACLAgentExecutor):

    def __init__(self, url):
        super().__init__(agentcard=my_card, my_url=url)
        print(
            "Warning: launching a pure python A2A agent,"
            " unable to check .asi interface against python body."
        )
        self.add_message_handler(Illocution.PROPOSE, self.execute_propose)
        self.spec = "A function to reverse the order of the elements in a list."

    async def execute_achieve(
        self,
        m: ACLMessage,
        output_event_queue: EventQueue,
    ):
        pass

    async def execute_tell(
        self,
        m: ACLIncomingMessage,
        output_event_queue: EventQueue,
    ):
        self.spec = m.content
        await sync_reply(output_event_queue, "ack")
        print("Received a new specification to consider.")

    async def execute_ask(
        self,
        m: ACLIncomingMessage,
        output_event_queue: EventQueue,
    ):
        pass

    async def execute_propose(
        self,
        m: ACLIncomingMessage,
        output_event_queue: EventQueue,
    ):

        print("(creating a new task)")
        task = new_task(m.origin.message)
        await output_event_queue.enqueue_event(task)
        updater = TaskUpdater(output_event_queue, task.id, task.context_id)

        await asyncio.sleep(1)

        await updater.update_status(
            TaskState.working,
            new_agent_text_message("start working", context_id=m.origin.context_id),
        )

        await asyncio.sleep(1)

        print("(consult llm)")
        try:
            res = ask_llm_for_correctness(
                self.spec,
                m.content,
            )
            await updater.update_status(
                TaskState.working,
                new_agent_text_message(
                    "finished working", context_id=m.origin.context_id
                ),
            )

            await asyncio.sleep(1)

            match res:
                case True:
                    atom = "valid"
                case False:
                    atom = "invalid"
                case None:
                    atom = "unknown"

            await updater.add_artifact(
                [Part(root=TextPart(text=str(atom)))], name="Final response."
            )

            await asyncio.sleep(1)

            await updater.complete()

            print("(replied " + str(atom) + " in an artifact)")

        except Exception:
            await updater.failed()
            print("Reported failure (synchronously)")



import context
from a2a_acl.agent.server_utils import run_server
from sample_agents.llm_validator_mistral_api.validator import (
    ValidatorAgentExecutor,
    my_card,
)
from a2a_acl.utils.url import build_url


the_host = "127.0.0.1"
my_url = build_url(the_host, context.validator_port)


async def main() -> None:
    run_server(
        ValidatorAgentExecutor(my_card, build_url(the_host, context.validator_port)),
        the_host,
        context.validator_port,
    )


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())



from a2a.types import AgentCard, AgentCapabilities

from a2a_acl.protocol_extension import EXTENSION
from a2a_acl.interface.interface import ACLAgentCard
from a2a_acl.interface.skill_conversion import (
    a2a_skill_of_bdi_skill,
    bdi_skill_of_a2a_skill,
)


def a2a_card_from_bdi_card(c: ACLAgentCard, url: str) -> AgentCard:
    asp_skills = [a2a_skill_of_bdi_skill(s) for s in c.skills]
    return AgentCard(
        name=c.name,
        description=c.doc,
        url=url,
        version="1.0.0",
        default_input_modes=["text"],
        default_output_modes=["text"],
        capabilities=AgentCapabilities(
            streaming=True,
            push_notifications=True,
            extensions=[EXTENSION],
        ),
        skills=asp_skills,
        supports_authenticated_extended_card=False,
    )


def bdi_card_from_a2a_card(c: AgentCard) -> ACLAgentCard:
    """
    Example of card for a BDI agent :
        additional_interfaces=None
        capabilities=AgentCapabilities(extensions=[AgentExtension(description='MOSAICO A2A BDI', params=None, required=True, uri='https://gitlab.eclipse.org/eclipse-research-labs/mosaico-project/a2a-agentspeak/-/blob/main/a2a_acl_protocol/MOSAICO_A2A_BDI_PROTOCOL')], push_notifications=True, state_transition_history=None, streaming=False)
        default_input_modes=['text']
        default_output_modes=['text']
        description='An agent which sends ping'
        documentation_url=None
        icon_url=None
        name='Pinger Agent'
        preferred_transport='JSONRPC'
        protocol_version='0.3.0'
        provider=None
        security=None
        security_schemes=None
        signatures=None
        skills=[
            AgentSkill(description='send a ping request (needs 1 parameter)', examples=['(achieve,do_ping(0))'], id='do_ping', input_modes=None, name='do_ping (from BDI agent)', output_modes=None, security=None, tags=['do_ping']),
            AgentSkill(description='send a secret (needs 1 parameter)', examples=['(achieve,share_secret(0))'], id='share_secret', input_modes=None, name='share_secret (from BDI agent)', output_modes=None, security=None, tags=['share_secret']),
            AgentSkill(description='do nothing (needs 0 parameter)', examples=['(tell,pong)'], id='pong', input_modes=None, name='pong (from BDI agent)', output_modes=None, security=None, tags=['pong'])
            ]
        supports_authenticated_extended_card=False
        url='http://127.0.0.1:9999/'
        version='1.0.0'
    """
    skills = [bdi_skill_of_a2a_skill(s) for s in c.skills]
    return ACLAgentCard(c.name, c.description, skills, [])



import re
from typing import Any

from a2a.server.agent_execution import RequestContext
from a2a.types import (
    SendMessageResponse,
    SendMessageSuccessResponse,
    Message,
)

from a2a_acl.protocol_extension import A2A_ACL_PROTOCOL_URI, MIME_TAG
from a2a_acl.protocol_extension.acl_message import ACLIncomingMessage


def check_illocution_regexpr(i: str) -> bool:
    """An illocution starts with a lowercase letter, and contains only lowercase letters and dash '-' character."""
    return bool(re.fullmatch(r"[a-z][a-z-]*", i))


def add_illocution(m: Message, i: str, codec: str) -> Message:
    """Add an illocution to a message with the ACL extension.
    The message is modified."""
    assert check_illocution_regexpr(i)
    if m.extensions is None:
        m.extensions = []
    if m.metadata is None:
        m.metadata = {}
    md: dict[str, Any] = {
        "illocution": i,
        "codec": codec,
        "mediaType": MIME_TAG,
    }
    m.extensions.append(A2A_ACL_PROTOCOL_URI)
    m.metadata[A2A_ACL_PROTOCOL_URI] = md
    return m


def extract_text(response: SendMessageResponse):
    """Extract text from synchronous replies"""
    if isinstance(response, SendMessageResponse):
        if isinstance(response.root, SendMessageSuccessResponse):
            if isinstance(response.root.result, Message):
                return response.root.result.parts[0].root.text
            else:
                print(
                    "Warning: Result of type: "
                    + str(type(response.root.result))
                    + " instead of Message."
                )
        else:
            print(
                "Warning: Root of type: "
                + str(type(response.root))
                + " instead of SendMessageSuccessResponse."
            )
    else:
        print(
            "Warning: Response of type: "
            + str(type(response))
            + " instead of SendMessageResponse."
        )

    return response.model_dump(mode="json", exclude_none=True)


def extract_illocution(context: RequestContext):
    return context.message.metadata[A2A_ACL_PROTOCOL_URI]["illocution"]


def extract_codec(context: RequestContext):
    return context.message.metadata[A2A_ACL_PROTOCOL_URI]["codec"]


def acl_of_a2a(context: RequestContext) -> ACLIncomingMessage:
    if (context.configuration is None) or (
        context.configuration.push_notification_config is None
    ):
        sender = None
    else:
        sender = context.configuration.push_notification_config.url
    if (
        context.message.metadata is None
        or A2A_ACL_PROTOCOL_URI not in context.message.metadata
    ):
        error_message = "Incoming message not in BDI format (missing metadata)."
        print(error_message)
        raise Exception(error_message)
    else:
        i = extract_illocution(context)
        c = extract_codec(context)
        content = context.get_user_input()

        return ACLIncomingMessage(
            i, content, sender, c, task_id=context.message.task_id, origin=context
        )



from abc import ABC, abstractmethod


class BadFormat(Exception):
    pass


class UnknownCodec(Exception):
    pass


class NotPublic(Exception):
    pass


class Codec(ABC):

    def __init__(self, name: str):
        self.name = name

    @abstractmethod
    def extract_entry_point(self, s: str):
        pass

    @abstractmethod
    def decode(self, s: str):
        pass


python_codec_id = "python_codec"
atom_codec_id = "atom_codec"
natural_language_id = "nl"



from a2a_acl.agent.server_utils import run_server
from sample_agents.ping.pingable import (
    PingableAgentExecutor,
    my_card,
)
from a2a_acl.utils.url import build_url

my_port = 9999
the_host = "127.0.0.1"
my_url = build_url(the_host, my_port)


async def main() -> None:
    run_server(
        PingableAgentExecutor(my_card, build_url(the_host, my_port)), the_host, my_port
    )


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())
    input("Press Enter to exit...\n")



from a2a_acl.interface.interface import (
    SkillDeclaration,
    ACLAgentCard,
)
from a2a_acl.protocol_extension.acl_message import Illocution


class SyntaxError(Exception):
    def __init__(self, message: str, filename: str, line: str):
        self.message = message
        self.filename = filename
        self.line = line


def read_file(intf: str) -> ACLAgentCard:
    with open(intf, "r") as f:
        l = f.readline()
        assert l.startswith("name = ")
        name = l.removeprefix("name = ").removesuffix("\n")

        l = f.readline()
        assert l.startswith("doc = ")
        agent_doc = l.removeprefix("doc = ").removesuffix("\n")

        lines = []

        for l in f:

            if l == "\n":
                pass
            else:
                try:
                    [k, functor, arity, doc] = l.removesuffix("\n").split(" : ")
                    match k:
                        case "tell" | "ask" | "achieve":
                            lines.append(
                                SkillDeclaration(
                                    declaration_kind=Illocution(k),
                                    doc=doc,
                                    functor=functor,
                                    arity=int(arity) if arity.isdigit() else None,
                                )
                            )
                        case "propose":
                            """proposal is an added illocution, with arity 1 (arity from file is ignored and can be a *)"""
                            lines.append(
                                SkillDeclaration(
                                    declaration_kind=Illocution.PROPOSE,
                                    doc=doc,
                                    functor=functor,
                                    arity=1,
                                )
                            )
                        case "upload":
                            """upload is an added illocution, with arity 1 (arity from file is ignored and can be a *)"""
                            lines.append(
                                SkillDeclaration(
                                    declaration_kind=Illocution.UPLOAD,
                                    doc=doc,
                                    functor=functor,
                                    arity=1,
                                )
                            )
                        case "failure":
                            """failure is an illocution from ACL, with arity 0 (arity from file is ignored and can be a *)"""
                            lines.append(
                                SkillDeclaration(
                                    declaration_kind=Illocution.FAILURE,
                                    doc=doc,
                                    functor=functor,
                                    arity=0,
                                )
                            )

                        case _:
                            raise SyntaxError(intf, l, "Bad illocution: " + k)
                except SyntaxError as e:
                    raise e
                except Exception:
                    raise SyntaxError(intf, l, "bad line structure")

        return ACLAgentCard(name, agent_doc, lines, [])



from a2a.types import AgentExtension

A2A_ACL_PROTOCOL_URI = (
    "https://gitlab.eclipse.org/eclipse-research-labs/mosaico-project/a2a-acl/"
)

EXTENSION = AgentExtension(
    uri=A2A_ACL_PROTOCOL_URI, description="MOSAICO A2A ACL", required=True
)

MIME_TAG = "application/x-mosaico-a2a-acl"



import logging

import httpx
from a2a.client import A2ACardResolver
from a2a.types import AgentCard
from a2a.utils.constants import AGENT_CARD_WELL_KNOWN_PATH


async def download_card(url: str) -> AgentCard:
    # logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(__name__)  # Get a logger instance

    async with httpx.AsyncClient(timeout=httpx.Timeout(timeout=30)) as httpx_client:
        # Initialize A2ACardResolver
        resolver = A2ACardResolver(
            httpx_client=httpx_client,
            base_url=url,
        )

        logger.info(
            f"Attempting to fetch public agent card from: {url}{AGENT_CARD_WELL_KNOWN_PATH}"
        )
        _public_card: AgentCard = (
            await resolver.get_agent_card()
        )  # Fetches from default public path
        return _public_card


class CardHolder:

    cards: list[AgentCard]
    reputation: dict[str, int]

    def __init__(self):
        self.cards = []
        self.reputation = {}

    async def retrieve_card_from(self, url) -> bool:
        logging.basicConfig(level=logging.INFO)
        logger = logging.getLogger(__name__)  # Get a logger instance

        try:
            _public_card: AgentCard = await download_card(url)

            logger.info("Successfully fetched public agent card:")
            self.cards.append(_public_card)
            logger.info(
                "\nUsing PUBLIC agent card for client initialization (default)."
            )
            return True

        except Exception as e:
            logger.info("Client failed to fetch the public agent card.")
            logger.error(e)
            return False

    def cards_with(self, predicate):
        return [c for c in self.cards if predicate(c)]

    def remove_by_url(self, url):
        self.cards = [c for c in self.cards if c.url != url]

    def get_reputation_by_url(self, url) -> int:
        match self.reputation.get(url):
            case None:
                return 0
            case other:
                return other

    def get_reputation_by_card(self, card):
        return self.get_reputation_by_url(card.url)

    def degrade(self, url: str):
        self.reputation[url] = self.get_reputation_by_url(url) - 1

    def get_card_by_url(self, url):
        for c in self.cards:
            if c.url == url:
                return c
        return None



from a2a.types import (
    AgentCard,
    Message,
)

from a2a_acl.a2a_utils.send_message import (
    build_message,
    SendMessageThread,
    default_message_handler,
    default_artifact_update_handler,
    default_status_update_handler,
    send_message,
    default_error_callback,
)
from a2a_acl.protocol_extension.message_codec import add_illocution


def build_acl_message(illocution: str, content: str, codec: str) -> Message:
    msg = build_message(content)
    msg = add_illocution(msg, illocution, codec)
    return msg


async def send_acl_message(
    destination: AgentCard | str,
    illocution: str,
    content: str,
    sender: str,
    codec: str,
    message_callback=default_message_handler,
    artifact_update_callback=default_artifact_update_handler,
    status_update_callback=default_status_update_handler,
) -> Message:
    """Send *text* to the target agent via the A2A ``message/send`` operation.

    Args:
        destination: card or url of the target agent.
        illocution : illocution such as 'tell' or 'achieve'
        content: Payload to send as a plain-text message.
        sender: the url of sender's server for asynchronous replies.
        codec: the codec used to code the content of the message.
        message_callback: the processor used to process the 'message' text parts.
        artifact_update_callback: the processor used to process the 'task' text parts.
        status_update_callback : the processor used to process the 'status' updates.

    Returns:
        The Message sent.
    """
    msg = build_acl_message(illocution, content, codec)

    await send_message(
        destination,
        msg,
        sender,
        message_callback,
        artifact_update_callback,
        status_update_callback,
    )
    return msg


def spawn_send_acl_message(
    dest: AgentCard | str,
    illocution: str,
    content: str,
    sender: str,
    codec: str,
    message_callback=default_message_handler,
    artifact_update_callback=default_artifact_update_handler,
    status_update_callback=default_status_update_handler,
    error_callback=default_error_callback,
) -> Message:
    msg = build_acl_message(illocution, content, codec)
    t = SendMessageThread(
        dest,
        msg,
        sender,
        error_callback,
        message_callback,
        artifact_update_callback,
        status_update_callback,
    )
    t.start()
    return msg
